from typing import TypeVar, Generic

NoneType = type(None)
byte = TypeVar("byte")
short = TypeVar("short")
long = TypeVar("long")
T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)


class FunctorLike(Generic[T_co]):
    __slots__ = ()



from datetime import date, datetime
from decimal import Decimal
from typing import List, Dict, Optional, Union, Callable, NamedTuple

import pytest
from pyspark.sql import types as t

from tinsel.lib import infer_spark_type, transform_field, maybe_unlift_optional, struct
from tinsel.types import NoneType, long, short, byte

DEFAULT_NAME = "some_field"


@struct
class Dummy(NamedTuple):
    pass


PRIMITIVES = [
    (NoneType, t.NullType()),
    (int, t.IntegerType()),
    (float, t.DoubleType()),
    (str, t.StringType()),
    (bytes, t.BinaryType()),
    (bytearray, t.BinaryType()),
    (bool, t.BooleanType()),
]

SYNTHETIC_PRIMITIVES = [
    (long, t.LongType()),
    (short, t.ShortType()),
    (byte, t.ByteType()),
]

DATE_TYPES = [
    (date, t.DateType()),
    (datetime, t.TimestampType()),
]

OPTIONALS_PY = [
    (str, (False, str)),
    (List[str], (False, List[str])),
    (Optional[str], (True, str)),
    (Optional[List[str]], (True, List[str])),
]

NULLABLE_TYPES = [
    (
        Optional[str],
        lambda name: t.StructField(name, t.StringType(), True)
    ),
    (
        Optional[List[str]],
        lambda name: t.StructField(name, t.ArrayType(t.StringType(), False), True)
    ),
    (
        List[Optional[str]],
        lambda name: t.StructField(name, t.ArrayType(t.StringType(), True), False)
    ),
    (
        Optional[List[Optional[str]]],
        lambda name: t.StructField(name, t.ArrayType(t.StringType(), True), True)
    ),
    (
        Optional[Dict[str, int]],
        lambda name: t.StructField(name, t.MapType(t.StringType(), t.IntegerType(), False), True)
    ),
    (
        Dict[str, Optional[int]],
        lambda name: t.StructField(name, t.MapType(t.StringType(), t.IntegerType(), True), False)
    ),
    (
        Optional[Dict[str, Optional[int]]],
        lambda name: t.StructField(name, t.MapType(t.StringType(), t.IntegerType(), True), True)
    ),
]


@pytest.mark.skip
def check(py_type, spark_type): assert infer_spark_type(py_type) == spark_type


@pytest.mark.parametrize(["py_type", "spark_type"], PRIMITIVES)
def test_primitive_types(py_type, spark_type): check(py_type, spark_type)


@pytest.mark.parametrize(["py_type", "spark_type"], SYNTHETIC_PRIMITIVES)
def test_synthetic_types(py_type, spark_type): check(py_type, spark_type)


@pytest.mark.parametrize(["py_type", "spark_type"], DATE_TYPES)
def test_date_types(py_type, spark_type): check(py_type, spark_type)


def test_decimal(): check(Decimal, t.DecimalType(precision=36, scale=6))


def test_list(): check(List[str], t.ArrayType(t.StringType(), False))


def test_dict(): check(Dict[int, str], t.MapType(t.IntegerType(), t.StringType(), False))


def test_struct(): check(Dummy, t.StructType([]))


def test_invalid_dict_keys():
    with pytest.raises(TypeError):
        infer_spark_type(Dict[Optional[int], str])


def test_invalid_type():
    with pytest.raises(TypeError):
        infer_spark_type(object)


def test_invalid_union():
    with pytest.raises(TypeError):
        infer_spark_type(Union[int, str])


def test_invalid_complex_type():
    with pytest.raises(TypeError):
        infer_spark_type(Callable[[int], str])


def test_field_transformer():
    assert transform_field(DEFAULT_NAME, str) == t.StructField(DEFAULT_NAME, t.StringType(), False)


@pytest.mark.parametrize(["raw", "expected"], OPTIONALS_PY)
def test_optionals_unlifting(raw, expected):
    assert maybe_unlift_optional(raw) == expected


@pytest.mark.parametrize(["raw", "field_factory"], NULLABLE_TYPES)
def test_nullable_fields(raw: type, field_factory: Callable[[str], t.StructField]):
    # noinspection PyTypeChecker
    assert transform_field(DEFAULT_NAME, raw) == field_factory(DEFAULT_NAME)



from datetime import date, datetime
from decimal import Decimal
from typing import Optional, Union, List, Tuple, Dict

from pyspark.sql import types as t

from tinsel.types import NoneType, byte, short, long, T, T_co, FunctorLike


def is_pyspark_class(cls: type) -> bool:
    return getattr(cls, "__pyspark_struct__", None) is ...


def is_container(cls: type) -> bool:
    fields = getattr(cls, "_fields", None)
    dc_fields = getattr(cls, "__dataclass_fields__", None)
    annotations = getattr(cls, "__annotations__", None)
    return (
        (isinstance(fields, tuple) or isinstance(dc_fields, dict))
        and isinstance(annotations, dict)
    )


def struct(cls: type) -> type:
    """
    Marks NamedTuple as suitable for ``tinsel.transform``

    :param cls: Typed NamedTuple
    """
    if not is_container(cls):
        raise ValueError(f"Only NamedTuple instances can be decorated with @struct, not {cls.__name__}")
    # Overwrite type with augmented one
    return type(cls.__name__, cls.__bases__, dict(__pyspark_struct__=..., **cls.__dict__))


def check_pyspark_struct(cls: type):
    if not isinstance(cls, type):
        raise TypeError(f"Expected type, but got instance {cls} of type {type(cls).__name__}")
    if is_container(cls):
        if not is_pyspark_class(cls):
            raise ValueError(f"Looks like type {cls.__name__} missed @struct decorator")
    else:
        raise ValueError(f"Type {cls.__name__} can't be used as structure")


def infer_nullability(typeclass) -> bool:
    import typing
    is_union = (
        isinstance(typeclass, getattr(typing, "_Union", type(T)))
        or getattr(typeclass, "__origin__", None) is Union
    )
    return is_union and NoneType in set(typeclass.__args__)


def unlift_optional(typeclass: Optional[T]) -> T:
    return list(set(typeclass.__args__) - {NoneType})[0]


def maybe_unlift_optional(typeclass: Union[T_co, FunctorLike[T_co]]) -> Tuple[bool, T_co]:
    is_nullable = infer_nullability(typeclass)
    return is_nullable, (unlift_optional(typeclass) if is_nullable else typeclass)


def infer_complex_spark_type(typeclass):
    if typeclass.__origin__ in {list, List}:
        co_T, *_ = typeclass.__args__
        is_nullable, py_type = maybe_unlift_optional(co_T)
        return t.ArrayType(infer_spark_type(py_type), is_nullable)
    elif typeclass.__origin__ in {dict, Dict}:
        k_T, v_T, *_ = typeclass.__args__
        is_nullable_key, py_key_type = maybe_unlift_optional(k_T)
        is_nullable_value, py_value_type = maybe_unlift_optional(v_T)
        if is_nullable_key:
            raise TypeError(f"Nullable keys of type {py_key_type} don't allowed in {typeclass}")
        return t.MapType(infer_spark_type(py_key_type), infer_spark_type(py_value_type), is_nullable_value)
    else:
        raise TypeError(f"Don't know how to represent {typeclass} in Spark")


def infer_spark_type(typeclass) -> t.DataType:
    if typeclass in (None, NoneType):
        return t.NullType()
    elif typeclass is str:
        return t.StringType()
    elif typeclass in {bytes, bytearray}:
        return t.BinaryType()
    elif typeclass is bool:
        return t.BooleanType()
    elif typeclass is date:
        return t.DateType()
    elif typeclass is datetime:
        return t.TimestampType()
    elif typeclass is Decimal:
        return t.DecimalType(precision=36, scale=6)
    elif typeclass is float:
        return t.DoubleType()
    elif typeclass is int:
        return t.IntegerType()
    elif typeclass is long:
        return t.LongType()
    elif typeclass is short:
        return t.ShortType()
    elif typeclass is byte:
        return t.ByteType()
    elif getattr(typeclass, "__origin__", None) is not None:
        return infer_complex_spark_type(typeclass)
    elif is_pyspark_class(typeclass):
        return transform(typeclass)
    else:
        raise TypeError(f"Don't know how to represent {typeclass} in Spark")


def transform_field(name: str, typeclass: type) -> t.StructField:
    (is_nullable, unwrapped_py_type) = maybe_unlift_optional(typeclass)
    return t.StructField(name, infer_spark_type(unwrapped_py_type), is_nullable)


def transform(typeclass: type) -> t.StructType:
    """
    Infer PySpark SQL types from namedtuple class fields

    Note: do not forget mark classes with ``@struct`` decorator!

    :param typeclass: @struct-annotated NamedTuple class
    :return: PySpark data structure
    """
    check_pyspark_struct(typeclass)
    return t.StructType([
        transform_field(name, cls)
        for name, cls
        in typeclass.__annotations__.items()
    ])



import sys
import pytest
from tinsel.lib import struct, transform
from pyspark.sql import types as t
from typing import NamedTuple, Optional, Dict, List

skip_dc = pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="Requires python3.7 or higher to test dataclasses"
)


@struct
class Plain(NamedTuple):
    a: str
    b: int
    c: bool


@struct
class Box(NamedTuple):
    d: float
    e: Plain
    f: None


@struct
class Complex(NamedTuple):
    g: Optional[Plain]
    h: List[Dict[str, Optional[Box]]]
    i: Dict[Box, Optional[List[Dict[int, Plain]]]]


COMPLEX_STRUCT = t.StructType([
    t.StructField(
        "g",
        t.StructType([
            t.StructField("a", t.StringType(), False),
            t.StructField("b", t.IntegerType(), False),
            t.StructField("c", t.BooleanType(), False)
        ]),
        True
    ),
    t.StructField(
        "h",
        t.ArrayType(
            t.MapType(
                t.StringType(),
                t.StructType([
                    t.StructField("d", t.DoubleType(), False),
                    t.StructField("e", t.StructType([
                        t.StructField("a", t.StringType(), False),
                        t.StructField("b", t.IntegerType(), False),
                        t.StructField("c", t.BooleanType(), False)
                    ]), False),
                    t.StructField("f", t.NullType(), False)
                ]), True),
            False
        ),
        False
    ),
    t.StructField(
        "i",
        t.MapType(
            t.StructType([
                t.StructField("d", t.DoubleType(), False),
                t.StructField("e", t.StructType([
                    t.StructField("a", t.StringType(), False),
                    t.StructField("b", t.IntegerType(), False),
                    t.StructField("c", t.BooleanType(), False)
                ]), False),
                t.StructField("f", t.NullType(), False)
            ]),
            t.ArrayType(
                t.MapType(
                    t.IntegerType(),
                    t.StructType([
                        t.StructField("a", t.StringType(), False),
                        t.StructField("b", t.IntegerType(), False),
                        t.StructField("c", t.BooleanType(), False)
                    ]),
                    False
                ),
                False
            ),
            True
        ),
        False
    )
])


def test_preserve_fields_order():
    assert tuple(transform(Plain).names) == Plain._fields


def test_parse_complex_type():
    assert transform(Complex) == COMPLEX_STRUCT


@skip_dc
def test_preserve_fields_order_dc():
    from dataclasses import dataclass

    @struct
    @dataclass
    class PlainDC:
        a: str
        b: int
        c: bool

    assert tuple(transform(PlainDC).names) == tuple(PlainDC.__dataclass_fields__.keys())


@skip_dc
def test_parse_complex_type_dc():
    from dataclasses import dataclass

    @struct
    @dataclass
    class PlainDC:
        a: str
        b: int
        c: bool

    @struct
    @dataclass
    class BoxDC:
        d: float
        e: PlainDC
        f: None

    @struct
    @dataclass
    class ComplexDC:
        g: Optional[PlainDC]
        h: List[Dict[str, Optional[BoxDC]]]
        i: Dict[Box, Optional[List[Dict[int, PlainDC]]]]

    assert transform(ComplexDC) == COMPLEX_STRUCT
    assert transform(Complex) == transform(ComplexDC)



import pytest

from collections import namedtuple
from typing import NamedTuple
from tinsel.lib import (
    struct,
    is_pyspark_class,
    is_container,
    check_pyspark_struct,
)


class Plain:
    pass


class Box(NamedTuple):
    pass


OldBox = namedtuple("OldBox", ["a"])


def test_struct_applicability():
    with pytest.raises(ValueError):
        struct(Plain)
    with pytest.raises(ValueError):
        struct(OldBox)
    assert is_pyspark_class(struct(Box))


def test_namedtuple_heuristics():
    assert is_container(Box)
    assert not is_container(OldBox)
    assert not is_container(tuple)


def test_pyspark_struct():
    AugmentedBox = struct(Box)
    with pytest.raises(TypeError):
        check_pyspark_struct(AugmentedBox())
    with pytest.raises(ValueError):
        check_pyspark_struct(Box)
    with pytest.raises(ValueError):
        check_pyspark_struct(Plain)



#!/usr/bin/env python
import sys
from pathlib import Path
from setuptools import setup, find_packages

if not sys.version_info >= (3, 6):
    raise RuntimeError('Tinsel requires at least python 3.6')

readme = Path('README.rst').read_text()
history = Path('HISTORY.rst').read_text()

requirements = ['pyspark>=2.3', ]
setup_requirements = ['pytest-runner', ]
test_requirements = ['pytest', ]

setup(
    author="Andriy Kushnir (Orhideous)",
    author_email='me@orhideous.name',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="PySpark schema generator",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='tinsel',
    name='tinsel',
    packages=find_packages(include=['tinsel']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/Orhideous/tinsel',
    version='0.2.0',
    zip_safe=False,
)
